﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using GameUtils.Graphics;
using GameUtils.Graphics.Data;
using SharpDX;
using SharpDX.Direct2D1;
using SharpDX.DirectWrite;
using FigureBegin = SharpDX.Direct2D1.FigureBegin;
using StrokeStyle = SharpDX.Direct2D1.StrokeStyle;
using Brush = SharpDX.Direct2D1.Brush;
using CapStyle = SharpDX.Direct2D1.CapStyle;
using LineJoin = SharpDX.Direct2D1.LineJoin;
using WordWrapping = SharpDX.DirectWrite.WordWrapping;

namespace GameUtils.Renderers.Direct2D1
{
    public class Direct2D1Renderer : Renderer
    {
        Control surface;
        SharpDX.Direct2D1.Factory d2dFactory;
        WindowRenderTarget target;
        SharpDX.DirectWrite.Factory textFactory;
        Direct2DFactory factory;
        bool initialized;
        Stack<SharpDX.Direct2D1.Layer> layers;

        protected override Graphics.Data.Factory Factory
        {
            get { return factory ?? (factory = new Direct2DFactory(d2dFactory, target, textFactory)); }
        }

        public override Version PlatformVersion
        {
            get { return new Version(6, 0); }
        }

        public override Math.Rectangle SurfaceBounds
        {
            get { return new Math.Rectangle(0, 0, surface.ClientSize.Width, surface.ClientSize.Height); }
        }

        protected override void Initialize(Control surface, PresentationOption presentationOption, AntiAliasMode antiAliasMode)
        {
            this.surface = surface;
            layers = new Stack<SharpDX.Direct2D1.Layer>();
            d2dFactory = new SharpDX.Direct2D1.Factory(SharpDX.Direct2D1.FactoryType.MultiThreaded);
            var props = new HwndRenderTargetProperties();
            props.Hwnd = surface.Handle;
            props.PixelSize = new Size2(surface.ClientSize.Width, surface.ClientSize.Height);
            props.PresentOptions = presentationOption == PresentationOption.Immediately ? PresentOptions.Immediately : PresentOptions.RetainContents;
            target = new WindowRenderTarget(d2dFactory, new RenderTargetProperties(), props);
            target.AntialiasMode = antiAliasMode == AntiAliasMode.AntiAlias ? AntialiasMode.PerPrimitive : AntialiasMode.Aliased;
            target.TextAntialiasMode = antiAliasMode == AntiAliasMode.AntiAlias ? SharpDX.Direct2D1.TextAntialiasMode.Cleartype : SharpDX.Direct2D1.TextAntialiasMode.Aliased;
            target.DotsPerInch = new Size2F(96, 96);
            textFactory = new SharpDX.DirectWrite.Factory();
            initialized = true;
        }

        protected override void BeginRender()
        {
            if (!initialized)
                throw new InvalidOperationException("Der Renderer muss initialisiert werden, bevor er verwendet werden kann.");
            var surfaceSize = new Size2(surface.ClientSize.Width, surface.ClientSize.Height);
            if (target.PixelSize != surfaceSize)
                target.Resize(surfaceSize);
            target.BeginDraw();
        }

        protected override void EndRender()
        {
            target.EndDraw();
        }

        private Brush GetD2DBrush<T>(Brush<T> brush) where T : BrushData
        {
            if (brush == null)
                throw new ArgumentNullException("brush");

            try
            {
                return ((IDirect2DBrush)this.GetResourceData(brush)).Brush;
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Brush-Objekt ist nicht kompatibel.", ex);
            }
        }

        private StrokeStyle GetD2DStrokeStyle(Graphics.StrokeStyle? strokeStyle)
        {
            if (!strokeStyle.HasValue)
                return null;

            var props = new StrokeStyleProperties();
            props.StartCap = (CapStyle)strokeStyle.Value.StartCap;
            props.EndCap = (CapStyle)strokeStyle.Value.EndCap;
            props.DashCap = (CapStyle)strokeStyle.Value.DashCap;
            props.DashOffset = strokeStyle.Value.DashOffset;
            props.MiterLimit = strokeStyle.Value.MiterLimit;
            props.LineJoin = (LineJoin)strokeStyle.Value.LineJoin;
            switch (strokeStyle.Value.DashStyle)
            {
                case Graphics.DashStyle.Custom:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.Custom;
                    break;
                case Graphics.DashStyle.Dash:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.Dash;
                    break;
                case Graphics.DashStyle.DashDot:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.DashDot;
                    break;
                case Graphics.DashStyle.DashDotDot:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.DashDotDot;
                    break;
                case Graphics.DashStyle.Dot:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.Dot;
                    break;
                default:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.Solid;
                    break;
            }

            if (strokeStyle.Value.DashStyle == Graphics.DashStyle.Custom)
                return new StrokeStyle(d2dFactory, props, strokeStyle.Value.DashPattern);
            else
                return new StrokeStyle(d2dFactory, props);
        }

        private void DrawLineInner<T>(Math.Vector2 p1, Math.Vector2 p2, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle? strokeStyle) where T : BrushData
        {
            Brush d2dBrush = GetD2DBrush(brush);
            using (StrokeStyle d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                target.DrawLine(Direct2DHelper.Vector2ToPointF(p1), Direct2DHelper.Vector2ToPointF(p2), d2dBrush, strokeWidth, d2dStrokeStyle);
            }
        }

        public override void DrawLine<T>(Math.Vector2 p1, Math.Vector2 p2, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawLineInner(p1, p2, brush, strokeWidth, null);
        }

        public override void DrawLine<T>(Math.Vector2 p1, Math.Vector2 p2, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle strokeStyle)
        {
            this.DrawLineInner(p1, p2, brush, strokeWidth, strokeStyle);
        }

        private void DrawEllipseInner<T>(Math.Ellipse ellipse, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle? strokeStyle) where T : BrushData
        {
            Brush d2dBrush = GetD2DBrush(brush);
            using (StrokeStyle d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                target.DrawEllipse(Direct2DHelper.GraphicsEllipseToD2DEllipse(ellipse), d2dBrush, strokeWidth, d2dStrokeStyle);
            }
        }

        public override void DrawEllipse<T>(Math.Ellipse ellipse, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawEllipseInner(ellipse, brush, strokeWidth, null);
        }

        public override void DrawEllipse<T>(Math.Ellipse ellipse, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle strokeStyle)
        {
            this.DrawEllipseInner(ellipse, brush, strokeWidth, strokeStyle);
        }

        public override void FillEllipse<T>(Math.Ellipse ellipse, Brush<T> brush)
        {
            Brush d2dBrush = GetD2DBrush(brush);
            target.FillEllipse(Direct2DHelper.GraphicsEllipseToD2DEllipse(ellipse), d2dBrush);
        }

        private void DrawRectangleInner<T>(Math.Rectangle rect, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle? strokeStyle) where T : BrushData
        {
            Brush d2dBrush = GetD2DBrush(brush);
            using (StrokeStyle d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                target.DrawRectangle(Direct2DHelper.ConvertRectangle(rect), d2dBrush, strokeWidth, d2dStrokeStyle);
            }
        }

        public override void DrawRectangle<T>(Math.Rectangle rect, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawRectangleInner(rect, brush, strokeWidth, null);
        }

        public override void DrawRectangle<T>(Math.Rectangle rect, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle strokeStyle)
        {
            this.DrawRectangleInner(rect, brush, strokeWidth, strokeStyle);
        }

        public override void FillRectangle<T>(Math.Rectangle rect, Brush<T> brush)
        {
            Brush d2dBrush = GetD2DBrush(brush);
            target.FillRectangle(Direct2DHelper.ConvertRectangle(rect), d2dBrush);
        }

        private void DrawPolygonInner<T>(Math.Polygon polygon, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle? strokeStyle) where T : BrushData
        {
            if (!polygon.IsValid)
                throw new ArgumentException("polygon", "Das übergebene Polygon ist nicht gültig.");

            Brush d2dBrush = GetD2DBrush(brush);
            using (StrokeStyle d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                using (var geometry = new PathGeometry(d2dFactory))
                {
                    using (var sink = geometry.Open())
                    {
                        sink.BeginFigure(Direct2DHelper.Vector2ToPointF(polygon.Points[0]), FigureBegin.Hollow);

                        for (int i = 1; i < polygon.Points.Length; i++)
                            sink.AddLine(Direct2DHelper.Vector2ToPointF(polygon.Points[i]));

                        sink.EndFigure(SharpDX.Direct2D1.FigureEnd.Closed);
                        sink.Close();

                        target.DrawGeometry(geometry, d2dBrush, strokeWidth, d2dStrokeStyle);
                    }
                }
            }
        }

        public override void DrawPolygon<T>(Math.Polygon polygon, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawPolygonInner(polygon, brush, strokeWidth, null);
        }

        public override void DrawPolygon<T>(Math.Polygon polygon, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle strokeStyle)
        {
            this.DrawPolygonInner(polygon, brush, strokeWidth, strokeStyle);
        }

        public override void FillPolygon<T>(Math.Polygon polygon, Brush<T> brush)
        {
            if (!polygon.IsValid)
                throw new ArgumentException("polygon", "Das übergebene Polygon ist nicht gültig.");

            Brush d2dBrush = GetD2DBrush(brush);
            using (var geometry = new PathGeometry(d2dFactory))
            {
                using (var sink = geometry.Open())
                {
                    sink.BeginFigure(Direct2DHelper.Vector2ToPointF(polygon.Points[0]), FigureBegin.Filled);

                    for (int i = 1; i < polygon.Points.Length; i++)
                        sink.AddLine(Direct2DHelper.Vector2ToPointF(polygon.Points[i]));

                    sink.EndFigure(SharpDX.Direct2D1.FigureEnd.Closed);
                    sink.Close();

                    target.FillGeometry(geometry, d2dBrush);
                }
            }
        }

        private void DrawTextInner<T>(string text, Graphics.Font font, Brush<T> brush, Math.Rectangle rect, Graphics.TextFormat format) where T : BrushData
        {
            Brush d2dBrush = GetD2DBrush(brush);

            if (font == null)
                throw new ArgumentNullException("font");
            Direct2DFont d2dFont;
            try
            {
                d2dFont = (Direct2DFont)this.GetResourceData(font);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Font-Objekt ist nicht kompatibel.", ex);
            }

            SharpDX.DirectWrite.TextFormat d2dFormat = d2dFont.GetTextFormat();
            if (format != null)
            {
                switch (format.HorizontalAlignment)
                {
                    case Graphics.HorizontalAlignment.Leading:
                        d2dFormat.TextAlignment = TextAlignment.Leading;
                        break;
                    case Graphics.HorizontalAlignment.Center:
                        d2dFormat.TextAlignment = TextAlignment.Center;
                        break;
                    case Graphics.HorizontalAlignment.Trailing:
                        d2dFormat.TextAlignment = TextAlignment.Trailing;
                        break;
                }
                switch (format.VerticalAlignment)
                {
                    case VerticalAlignment.Top:
                        d2dFormat.ParagraphAlignment = ParagraphAlignment.Near;
                        break;
                    case VerticalAlignment.Center:
                        d2dFormat.ParagraphAlignment = ParagraphAlignment.Center;
                        break;
                    case VerticalAlignment.Bottom:
                        d2dFormat.ParagraphAlignment = ParagraphAlignment.Far;
                        break;
                }
                d2dFormat.WordWrapping = (format.WordWrapping == Graphics.WordWrapping.Wrap) ? WordWrapping.Wrap : WordWrapping.NoWrap;
            }
            else
            {
                d2dFormat.TextAlignment = TextAlignment.Leading;
                d2dFormat.ParagraphAlignment = ParagraphAlignment.Near;
                d2dFormat.WordWrapping = WordWrapping.Wrap;
            }

            DrawTextOptions options = format == null || format.WordWrapping != Graphics.WordWrapping.Clip ? DrawTextOptions.NoSnap : DrawTextOptions.Clip | DrawTextOptions.NoSnap;
            target.DrawText(text, d2dFormat, Direct2DHelper.ConvertRectangle(rect), d2dBrush, options);
        }

        public override void DrawText<T>(string text, Graphics.Font font, Brush<T> brush, Math.Rectangle rect, Graphics.TextFormat format = null)
        {
            this.DrawTextInner(text, font, brush, rect, format);
        }

        public override void DrawText<T>(string text, Graphics.Font font, Brush<T> brush, Math.Vector2 pt)
        {
            this.DrawTextInner(text, font, brush, new Math.Rectangle(pt.X, pt.Y, float.MaxValue, float.MaxValue), null);
        }

        private void DrawTextureInner(Texture texture, Math.Rectangle destRect, Math.Rectangle? sourceRect, float opacity, Graphics.InterpolationMode interpolationMode)
        {
            if (texture == null)
                throw new ArgumentNullException("texture");
            Direct2DTexture d2dTexture;
            try
            {
                d2dTexture = (Direct2DTexture)this.GetResourceData(texture);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Diese Texture ist nicht kompatibel.", ex);
            }
            Bitmap d2dBitmap = d2dTexture.GetBitmap();

            target.DrawBitmap(d2dBitmap, Direct2DHelper.ConvertRectangle(destRect), opacity,
                interpolationMode == Graphics.InterpolationMode.NearestNeighbor ? BitmapInterpolationMode.NearestNeighbor : BitmapInterpolationMode.Linear,
                sourceRect.HasValue ? Direct2DHelper.ConvertRectangle(sourceRect.Value) : (RectangleF?)null);
        }

        public override void DrawTexture(Texture texture, Math.Rectangle destRect, Math.Rectangle? sourceRect = null, float opacity = 1.0f, Graphics.InterpolationMode interpolationMode = Graphics.InterpolationMode.HighQualityCubic)
        {
            this.DrawTextureInner(texture, destRect, sourceRect, opacity, interpolationMode);
        }

        public override void DrawTexture(Texture texture, Math.Vector2 pt, float opacity = 1.0f)
        {
            this.DrawTextureInner(texture, new Math.Rectangle(pt.X, pt.Y, texture.Width, texture.Height), null, opacity, Graphics.InterpolationMode.HighQualityCubic);
        }

        private void DrawGeometryInner<T>(Graphics.Geometry geometry, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle? strokeStyle) where T : BrushData
        {
            Brush d2dBrush = GetD2DBrush(brush);

            if (geometry == null)
                throw new ArgumentNullException("geometry");
            Direct2DGeometry d2dGeometry;
            try
            {
                d2dGeometry = (Direct2DGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            using (StrokeStyle d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                target.DrawGeometry(d2dGeometry.Geometry, d2dBrush, strokeWidth, d2dStrokeStyle);
            }
        }

        public override void DrawGeometry<T>(Graphics.Geometry geometry, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawGeometryInner(geometry, brush, strokeWidth, null);
        }

        public override void DrawGeometry<T>(Graphics.Geometry geometry, Brush<T> brush, float strokeWidth, Graphics.StrokeStyle strokeStyle)
        {
            this.DrawGeometryInner(geometry, brush, strokeWidth, strokeStyle);
        }

        public override void FillGeometry<T>(Graphics.Geometry geometry, Brush<T> brush)
        {
            Brush d2dBrush = GetD2DBrush(brush);

            if (geometry == null)
                throw new ArgumentNullException("geometry");
            Direct2DGeometry d2dGeometry;
            try
            {
                d2dGeometry = (Direct2DGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            target.FillGeometry(d2dGeometry.Geometry, d2dBrush);
        }

        public override void PushClip(Graphics.Geometry geometry)
        {
            if (geometry == null)
                throw new ArgumentNullException("geometry");
            Direct2DGeometry d2dGeometry;
            try
            {
                d2dGeometry = (Direct2DGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            var layer = new SharpDX.Direct2D1.Layer(target);
            layers.Push(layer);
            var layerParams = new LayerParameters();
            layerParams.GeometricMask = d2dGeometry.Geometry;
            layerParams.MaskAntialiasMode = AntialiasMode.PerPrimitive;
            layerParams.LayerOptions = LayerOptions.None;
            layerParams.Opacity = 1;
            target.PushLayer(ref layerParams, layer);
        }

        public override void PushClip(Math.Rectangle rect)
        {
            var layer = new SharpDX.Direct2D1.Layer(target);
            layers.Push(layer);
            var layerParams = new LayerParameters();
            layerParams.ContentBounds = Direct2DHelper.ConvertRectangle(rect);
            layerParams.LayerOptions = LayerOptions.None;
            layerParams.Opacity = 1;
            target.PushLayer(ref layerParams, layer);
        }
        
        public override void PopClip()
        {
            target.PopLayer();
            layers.Pop().Dispose();
        }

        public override void Clear(Graphics.Color4 color)
        {
            target.Clear(Direct2DHelper.GraphicsColorToD2DColor(color));
        }

        public override void SetTransform(Math.Matrix2x3 matrix)
        {
            var d2dMatrix = new Matrix3x2();
            d2dMatrix.TranslationVector = new Vector2(matrix.OffsetX, matrix.OffsetY);
            d2dMatrix.ScaleVector = new Vector2(matrix[0, 0], matrix[1, 1]);
            d2dMatrix.M12 = matrix[1, 0];
            d2dMatrix.M21 = matrix[0, 1];
            target.Transform = d2dMatrix;
        }

        public override void ResetTransform()
        {
            target.Transform = Matrix3x2.Identity;
        }

        public override Math.Vector2 MeasureText(string text, Graphics.Font font)
        {
            if (font == null)
                throw new ArgumentNullException("font");
            Direct2DFont d2dFont;
            try
            {
                d2dFont = (Direct2DFont)this.GetResourceData(font);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Font-Objekt ist nicht kompatibel.", ex);
            }
            SharpDX.DirectWrite.TextFormat d2dFormat = d2dFont.GetTextFormat();
            d2dFormat.ParagraphAlignment = ParagraphAlignment.Near;
            d2dFormat.TextAlignment = TextAlignment.Leading;
            d2dFormat.WordWrapping = WordWrapping.NoWrap;

            using (var layout = new TextLayout(textFactory, text, d2dFormat, float.MaxValue, float.MaxValue))
            {
                return new Math.Vector2(layout.Metrics.Width, layout.Metrics.Height);
            }
        }

        protected override void Dispose(bool disposing)
        {
            while (layers != null && layers.Count > 0)
                layers.Pop().Dispose();
            if (disposing)
            {
                factory = null;
                layers = null;
            }
            if (textFactory != null) textFactory.Dispose();
            if (target != null) target.Dispose();
            if (d2dFactory != null) d2dFactory.Dispose();
        }
    }
}
