﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Text;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using GameUtils.Graphics;
using GameUtils.Graphics.Data;
using GameUtils.Math;
using Brush = System.Drawing.Brush;

namespace GameUtils.Renderers.Gdi
{
    public class GdiRenderer : Renderer
    {
        GdiFactory factory;
        Control surface;
        bool rendering;
        Bitmap bmp;
        System.Drawing.Graphics device;
        bool initialized;
        bool enableVSync;
        bool enableAntiAlias;
        Stopwatch sw;
        Stack<GraphicsPath> layers;
        Matrix transformationMatrix;
        bool resized;
        bool forceImmediatelyDraw;

        [DllImport("gdi32.dll")]
        private static extern bool BitBlt(IntPtr hObject, int nXDest, int nYDest, int nWidth, int nHeight, IntPtr hObjSource, int nXSrc, int nYSrc, TernaryRasterOperations dwRop);

        [DllImport("gdi32.dll", ExactSpelling = true, SetLastError = true)]
        private static extern IntPtr CreateCompatibleDC(IntPtr hdc);

        [DllImport("gdi32.dll", ExactSpelling = true, SetLastError = true)]
        private static extern bool DeleteDC(IntPtr hdc);

        [DllImport("gdi32.dll", ExactSpelling = true, SetLastError = true)]
        private static extern IntPtr SelectObject(IntPtr hdc, IntPtr hgdiobj);

        [DllImport("gdi32.dll", ExactSpelling = true, SetLastError = true)]
        private static extern bool DeleteObject(IntPtr hObject);

        private enum TernaryRasterOperations : uint
        {
            Srccopy = 0x00CC0020,
        }

        protected override Factory Factory
        {
            get { return factory ?? (factory = new GdiFactory()); }
        }

        public override Version PlatformVersion
        {
            get { return new Version(5, 1); }
        }

        public override Math.Rectangle SurfaceBounds
        {
            get { return new Math.Rectangle(0, 0, surface.ClientSize.Width, surface.ClientSize.Height); }
        }

        private void CreateDevice()
        {
            if (bmp != null) bmp.Dispose();
            bmp = new Bitmap(surface.ClientSize.Width, surface.ClientSize.Height, PixelFormat.Format32bppArgb);
            device = System.Drawing.Graphics.FromImage(bmp);
            if (enableAntiAlias)
            {
                device.SmoothingMode = SmoothingMode.AntiAlias;
                device.PixelOffsetMode = PixelOffsetMode.HighQuality;
                device.TextRenderingHint = TextRenderingHint.ClearTypeGridFit;
            }
            else
            {
                device.SmoothingMode = SmoothingMode.HighSpeed;
                device.PixelOffsetMode = PixelOffsetMode.HighSpeed;
                device.TextRenderingHint = TextRenderingHint.SystemDefault;
            }
            if (transformationMatrix != null) device.Transform = transformationMatrix;
        }

        protected override void Initialize(Control surface, PresentationOption presentationOption, AntiAliasMode antiAliasMode)
        {
            this.surface = surface;
            surface.Resize += (sender, e) => resized = true;
            layers = new Stack<GraphicsPath>();
            initialized = true;
            if (presentationOption == PresentationOption.Limited)
            {
                enableVSync = true;
                sw = new Stopwatch();
            }
            enableAntiAlias = antiAliasMode == AntiAliasMode.AntiAlias;
            this.CreateDevice();
        }

        protected override void BeginRender()
        {
            if (!initialized) throw new InvalidOperationException("Der Renderer muss initialisiert werden, bevor er verwendet werden kann.");
            rendering = true;
        }

        protected override void EndRender()
        {
            rendering = false;
            if (enableVSync)
            {
                if (sw.IsRunning && !resized && !forceImmediatelyDraw)
                {
                    if (sw.ElapsedMilliseconds < 15)
                        Thread.Sleep(System.Math.Max(15 - (int)sw.ElapsedMilliseconds, 0));
                    while (sw.Elapsed.TotalMilliseconds < 1000.0 / 60.0) { }
                }
                forceImmediatelyDraw = false;
                sw.Restart();
            }
            if (surface.IsHandleCreated)
            {
                var g = surface.CreateGraphics();
                IntPtr pTarget = g.GetHdc();
                IntPtr pSource = CreateCompatibleDC(pTarget);
                IntPtr hBmp = bmp.GetHbitmap();
                IntPtr pOrig = SelectObject(pSource, hBmp);
                BitBlt(pTarget, 0, 0, bmp.Width, bmp.Height, pSource, 0, 0, TernaryRasterOperations.Srccopy);
                DeleteObject(hBmp);
                DeleteDC(pSource);
                g.ReleaseHdc(pTarget);
                g.Dispose();
            }

            if (resized)
            {
                this.CreateDevice();
                forceImmediatelyDraw = true;
                resized = false;
            }
        }

        private void CheckRenderState()
        {
            if (!rendering)
                throw new InvalidOperationException("Zeichenvorgänge sind nur zwischen \"BeginRender\" und \"EndRender\" gültig.");
        }

        private Brush ConvertBrush<T>(Brush<T> brush) where T : BrushData
        {
            if (brush == null)
                throw new ArgumentNullException("brush");

            try
            {
                return ((IGdiBrush)this.GetResourceData(brush)).Brush;
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Brush-Objekt ist nicht kompatibel.", ex);
            }
        }

        private Pen GetPen(Brush brush, float strokeWidth, StrokeStyle? strokeStyle)
        {
            var pen = new Pen(brush, strokeWidth);
            if (strokeStyle.HasValue)
            {
                pen.StartCap = GdiHelper.GetLineCap(strokeStyle.Value.StartCap);
                pen.EndCap = GdiHelper.GetLineCap(strokeStyle.Value.EndCap);
                pen.DashCap = GdiHelper.GetDashCap(strokeStyle.Value.DashCap);
                pen.DashOffset = strokeStyle.Value.DashOffset;
                pen.DashPattern = strokeStyle.Value.DashPattern;
                pen.LineJoin = (System.Drawing.Drawing2D.LineJoin)strokeStyle.Value.LineJoin;
                pen.MiterLimit = strokeStyle.Value.MiterLimit;
                pen.DashStyle = GdiHelper.GetDashStyle(strokeStyle.Value.DashStyle);
            }
            return pen;
        }

        private void DrawLineInner<T>(Vector2 p1, Vector2 p2, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            CheckRenderState();

            Brush gdiBrush = ConvertBrush(brush);
            using (Pen p = this.GetPen(gdiBrush, strokeWidth, strokeStyle))
            {
                device.DrawLine(p, GdiHelper.Vector2ToPointF(p1), GdiHelper.Vector2ToPointF(p2));
            }
        }

        public override void DrawLine<T>(Vector2 p1, Vector2 p2, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawLineInner(p1, p2, brush, strokeWidth, null);
        }

        public override void DrawLine<T>(Vector2 p1, Vector2 p2, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawLineInner(p1, p2, brush, strokeWidth, strokeStyle);
        }

        private void DrawEllipseInner<T>(Ellipse ellipse, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            CheckRenderState();

            Brush gdiBrush = ConvertBrush(brush);
            using (Pen p = this.GetPen(gdiBrush, strokeWidth, strokeStyle))
            {
                device.DrawEllipse(p, GdiHelper.GetEllipseRect(ellipse));
            }
        }

        public override void DrawEllipse<T>(Ellipse ellipse, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawEllipseInner(ellipse, brush, strokeWidth, null);
        }

        public override void DrawEllipse<T>(Ellipse ellipse, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawEllipseInner(ellipse, brush, strokeWidth, strokeStyle);
        }

        public override void FillEllipse<T>(Ellipse ellipse, Brush<T> brush)
        {
            CheckRenderState();

            var gdiBrush = ConvertBrush(brush);	
            device.FillEllipse(gdiBrush, GdiHelper.GetEllipseRect(ellipse));
        }

        private void DrawRectangleInner<T>(Math.Rectangle rect, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            CheckRenderState();

            Brush gdiBrush = ConvertBrush(brush);
            using (Pen p = this.GetPen(gdiBrush, strokeWidth, strokeStyle))
            {
                device.DrawRectangle(p, rect.Location.X, rect.Location.Y, rect.Size.X, rect.Size.Y);
            }
        }

        public override void DrawRectangle<T>(Math.Rectangle rect, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawRectangleInner(rect, brush, strokeWidth, null);
        }

        public override void DrawRectangle<T>(Math.Rectangle rect, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawRectangleInner(rect, brush, strokeWidth, null);
        }

        public override void FillRectangle<T>(Math.Rectangle rect, Brush<T> brush)
        {
            CheckRenderState();

            var gdiBrush = ConvertBrush(brush);	
            device.FillRectangle(gdiBrush, GdiHelper.ConvertRectangle(rect));
        }

        private void DrawPolygonInner<T>(Polygon polygon, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            CheckRenderState();

            Brush gdiBrush = ConvertBrush(brush);
            using (Pen p = this.GetPen(gdiBrush, strokeWidth, strokeStyle))
            {
                var points = new PointF[polygon.Points.Length];
                for (int i = 0; i < points.Length; i++)
                    points[i] = new PointF(polygon.Points[i].X, polygon.Points[i].Y);
                device.DrawPolygon(p, points);
            }
        }

        public override void DrawPolygon<T>(Polygon polygon, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawPolygonInner(polygon, brush, strokeWidth, null);
        }

        public override void DrawPolygon<T>(Polygon polygon, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawPolygonInner(polygon, brush, strokeWidth, null);
        }

        public override void FillPolygon<T>(Polygon polygon, Brush<T> brush)
        {
            CheckRenderState();

            var gdiBrush = ConvertBrush(brush);
            var points = new PointF[polygon.Points.Length];
            for (int i = 0; i < points.Length; i++)
                points[i] = new PointF(polygon.Points[i].X, polygon.Points[i].Y);
            device.FillPolygon(gdiBrush, points);
        }

        private void DrawTextInner<T>(string text, Graphics.Font font, Brush<T> brush, Math.Rectangle rect, TextFormat format) where T : BrushData
        {
            CheckRenderState();

            GdiFont gdiFont;
            try
            {
                gdiFont = (GdiFont)this.GetResourceData(font);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Font-Objekt ist nicht kompatibel.", ex);
            }
            var gdiBrush = ConvertBrush(brush);

            StringFormat gdiFormat;
            if (format != null)
            {
                try
                {
                    gdiFormat = ((GdiTextFormat)this.GetResourceData(format)).GetStringFormat();
                }
                catch (InvalidCastException ex)
                {
                    throw new ArgumentException("Dieses TextFormat-Objekt ist nicht kompatibel.", ex);
                }
            }
            else
            {
                gdiFormat = StringFormat.GenericDefault;
            }

            if (font.Size > 21)
            {
                using (var path = new GraphicsPath())
                {
                    path.AddString(text, gdiFont.GetFont().FontFamily, (int)gdiFont.GetFont().Style, gdiFont.GetFont().Size, GdiHelper.ConvertRectangle(rect), gdiFormat);
                    device.FillPath(gdiBrush, path);
                }
            }
            else
            {
                device.DrawString(text, gdiFont.GetFont(), gdiBrush, GdiHelper.ConvertRectangle(rect), gdiFormat);
            }
        }

        public override void DrawText<T>(string text, Graphics.Font font, Brush<T> brush, Math.Rectangle rect, TextFormat format = null)
        {
            this.DrawTextInner(text, font, brush, rect, format);
        }

        public override void DrawText<T>(string text, Graphics.Font font, Brush<T> brush, Vector2 pt)
        {
            this.DrawTextInner(text, font, brush, new Math.Rectangle(pt.X, pt.Y, float.MaxValue, float.MaxValue), null);
        }

        private void DrawTextureInner(Texture texture, Math.Rectangle destRect, Math.Rectangle? sourceRect, float opacity, Graphics.InterpolationMode interpolationMode)
        {
            CheckRenderState();

            GdiTexture gdiBmp;
            try
            {
                gdiBmp = (GdiTexture)this.GetResourceData(texture);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Diese Texture ist nicht kompatibel.", ex);
            }

            device.InterpolationMode = GdiHelper.ConvertInterpolytionMode(interpolationMode);

            opacity = MathHelper.Clamp(opacity, 0, 1);
            if (opacity == 1.0f)
            {
                device.DrawImage(gdiBmp.GetBitmap(), GdiHelper.ConvertRectangle(destRect),
                    sourceRect.HasValue ? GdiHelper.ConvertRectangle(sourceRect.Value) : new RectangleF(0, 0, texture.Width, texture.Height), GraphicsUnit.Pixel);
            }
            else
            {
                var colorMatrix = new ColorMatrix();
                colorMatrix.Matrix33 = opacity;
                using (var imageAttributes = new ImageAttributes())
                {
                    imageAttributes.SetColorMatrix(colorMatrix, ColorMatrixFlag.Default, ColorAdjustType.Bitmap);
                    device.DrawImage(gdiBmp.GetBitmap(), new[] { GdiHelper.Vector2ToPointF(destRect.Location), new PointF(destRect.X + destRect.Width, destRect.Y),
                        new PointF(destRect.X + destRect.Width, destRect.Y + destRect.Height) },
                        sourceRect.HasValue ? GdiHelper.ConvertRectangle(sourceRect.Value) : new RectangleF(0, 0, texture.Width, texture.Height),
                        GraphicsUnit.Pixel, imageAttributes);
                }
            }
        }

        public override void DrawTexture(Texture texture, Math.Rectangle destRect, Math.Rectangle? sourceRect, float opacity = 1.0f, Graphics.InterpolationMode interpolationMode = Graphics.InterpolationMode.HighQualityCubic)
        {
            this.DrawTextureInner(texture, destRect, null, opacity, interpolationMode);
        }

        public override void DrawTexture(Texture texture, Vector2 pt, float opacity = 1.0f)
        {
            this.DrawTextureInner(texture, new Math.Rectangle(pt.X, pt.Y, texture.Width, texture.Height), null, opacity, Graphics.InterpolationMode.HighQualityCubic);
        }

        private void DrawGeometryInner<T>(Geometry geometry, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            CheckRenderState();

            Brush gdiBrush = ConvertBrush(brush);
            GdiGeometry gdiPath;
            try
            {
                gdiPath = (GdiGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            using (Pen p = this.GetPen(gdiBrush, strokeWidth, strokeStyle))
            {
                device.DrawPath(p, gdiPath.Path);
            }
        }

        public override void DrawGeometry<T>(Geometry geometry, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawGeometryInner(geometry, brush, strokeWidth, null);
        }

        public override void DrawGeometry<T>(Geometry geometry, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawGeometryInner(geometry, brush, strokeWidth, strokeStyle);
        }

        public override void FillGeometry<T>(Geometry geometry, Brush<T> brush)
        {
            CheckRenderState();

            var gdiBrush = ConvertBrush(brush);
            GdiGeometry gdiPath;
            try
            {
                gdiPath = (GdiGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            device.FillRegion(gdiBrush, gdiPath.Region);
        }

        public override void PushClip(Geometry geometry)
        {
            CheckRenderState();

            GdiGeometry gdiPath;
            try
            {
                gdiPath = (GdiGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            device.SetClip(gdiPath.Path);
            layers.Push(gdiPath.Path);
        }

        public override void PushClip(Math.Rectangle rect)
        {
            CheckRenderState();

            var gdiPath = new GraphicsPath();
            gdiPath.AddRectangle(GdiHelper.ConvertRectangle(rect));
            device.SetClip(gdiPath);
            layers.Push(gdiPath);
        }

        public override void PopClip()
        {
            CheckRenderState();

            if (layers.Count > 0)
            {
                device.ResetClip();
                layers.Pop();
                if (layers.Count > 0) device.SetClip(layers.Peek());
            }
            else
                throw new InvalidOperationException("Zu viele Pop-Aufrufe.");
        }

        public override void Clear(Color4 color)
        {
            CheckRenderState();

            device.Clear(GdiHelper.ConvertColor(color));
        }

        public override void SetTransform(Matrix2x3 matrix)
        {
            var gdiMatrix = new Matrix(matrix[0, 0], matrix[1, 0], matrix[0, 1], matrix[1, 1], matrix.OffsetX, matrix.OffsetY);
            if (device != null) device.Transform = gdiMatrix;
            if (transformationMatrix != null) transformationMatrix.Dispose();
            transformationMatrix = gdiMatrix;
        }

        public override void ResetTransform()
        {
            if (device != null) device.ResetTransform();
            if (transformationMatrix != null) transformationMatrix.Dispose();
            transformationMatrix = null;
        }

        public override Vector2 MeasureText(string text, Graphics.Font font)
        {
            GdiFont gdiFont;
            try
            {
                gdiFont = (GdiFont)this.GetResourceData(font);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Font-Objekt ist nicht kompatibel.", ex);
            }

            Size size = TextRenderer.MeasureText(text, gdiFont.GetFont());
            return new Vector2(size.Width, size.Height);
        }

        protected override void Dispose(bool disposing)
        {
            if(disposing)
            {
                factory = null;
                layers = null;
            }
            if (device != null) device.Dispose();
            if (bmp != null) bmp.Dispose();
            if (transformationMatrix != null) transformationMatrix.Dispose();
        }
    }
}
