﻿using System;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using GameUtils.Graphics.Data;
using Drawing = System.Drawing;
using GameUtils.Graphics;
using WrapMode = GameUtils.Graphics.TextureWrapMode;

namespace GameUtils.Renderers.Gdi
{
    internal class GdiTextureBrush : TextureBrushData, IGdiBrush
    {
        Drawing.TextureBrush brush;
        Texture texture;
        Drawing.Bitmap bmp;
        float opacity;
        Matrix matrix;

        public Drawing.Brush Brush
        {
            get { return brush; }
        }

        protected override Texture Texture
        {
            get { return texture; }
            set
            {
                if (value == null)
                    throw new ArgumentNullException("value");

                GdiTexture data;
                try
                {
                    data = (GdiTexture)this.GetResourceData(value);
                }
                catch (InvalidCastException ex)
                {
                    throw new ArgumentException("Diese Texture ist nicht kompatibel.", ex);
                }

                this.texture = value;
                bmp = new Drawing.Bitmap(data.GetBitmap().Width, data.GetBitmap().Height);
                brush.Dispose();
                bmp.Dispose();
                brush = CreateBrush();
            }
        }

        protected override WrapMode WrapMode
        {
            get
            {
                switch (brush.WrapMode)
                {
                    case Drawing.Drawing2D.WrapMode.Clamp:
                        return TextureWrapMode.None;
                    case Drawing.Drawing2D.WrapMode.Tile:
                        return TextureWrapMode.Tile;
                    default:
                        return TextureWrapMode.Mirror;
                }
            }
            set
            {
                switch (value)
                {
                    case TextureWrapMode.None:
                        brush.WrapMode = Drawing.Drawing2D.WrapMode.Clamp;
                        break;
                    case TextureWrapMode.Tile:
                        brush.WrapMode = Drawing.Drawing2D.WrapMode.Tile;
                        break;
                    default:
                        brush.WrapMode = Drawing.Drawing2D.WrapMode.TileFlipXY;
                        break;
                }
            }
        }

        protected override float Opacity
        {
            get { return opacity; }
            set
            {
                if (value > 1)
                    opacity = 1;
                else if (value < 0)
                    opacity = 0;
                else
                    opacity = value;
                brush.Dispose();
                bmp.Dispose();
                brush = CreateBrush();
            }
        }

        private Drawing.TextureBrush CreateBrush()
        {
            var colorMatrix = new ColorMatrix();
            colorMatrix.Matrix33 = opacity;
            var imageAttributes = new ImageAttributes();
            imageAttributes.SetColorMatrix(colorMatrix, ColorMatrixFlag.Default, ColorAdjustType.Bitmap);
            using (Drawing.Graphics g = Drawing.Graphics.FromImage(bmp))
            {
                g.DrawImage(((GdiTexture)this.GetResourceData(texture)).GetBitmap(), new Drawing.Rectangle(0, 0, bmp.Width, bmp.Height), 0, 0, bmp.Width, bmp.Height, Drawing.GraphicsUnit.Pixel, imageAttributes);
            }
            var b = new Drawing.TextureBrush(bmp);
            if (matrix != null) b.Transform = matrix;
            return b;
        }

        internal GdiTextureBrush(Texture texture)
        {
            GdiTexture data;
            try
            {
                data = (GdiTexture)this.GetResourceData(texture);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Diese Texture ist nicht kompatibel.", ex);
            }

            this.texture = texture;
            bmp = new Drawing.Bitmap(data.GetBitmap().Width, data.GetBitmap().Height);
            opacity = 1;
            brush = CreateBrush();
        }

        protected override void SetTransform(Math.Matrix2x3 matrix)
        {
            var gdiMatrix = new Matrix(matrix[0, 0], matrix[1, 0], matrix[0, 1], matrix[1, 1], matrix.OffsetX, matrix.OffsetY);
            brush.Transform = gdiMatrix;
            if (this.matrix != null)
                this.matrix.Dispose();
            this.matrix = gdiMatrix;
        }

        protected override void ResetTransform()
        {
            brush.ResetTransform();
            if (matrix != null)
                matrix.Dispose();
            matrix = null;
        }

        protected override void Dispose(bool disposing)
        {
            brush.Dispose();
            bmp.Dispose();
            if (matrix != null) matrix.Dispose();
        }
    }
}
