﻿using System;
using System.Windows.Forms;

namespace GameUtils.Graphics
{
    /// <summary>
    /// The game window provides basic fuctionality for a proper renderer surface.
    /// </summary>
    public abstract class GameWindowBase : Form
    {
        readonly bool disableCloseButton;
        bool fullscreen;
        FormBorderStyle oldBorderStyle;
        FormWindowState oldWindowState;

        /// <summary>
        /// Is risen when the fullscreen mode has changed.
        /// </summary>
        public event EventHandler FullscreenChanged;

        protected override CreateParams CreateParams
        {
            get
            {
                CreateParams param = base.CreateParams;
                if (disableCloseButton) param.ClassStyle = param.ClassStyle | 0x200;
                return param;
            }
        }

        /// <summary>
        /// Indicates whether the window is in fullscreen mode.
        /// </summary>
        public bool Fullscreen
        {
            get { return fullscreen; }
            set
            {
                if (value != fullscreen)
                {
                    var action = new Action(() =>
                    {
                        fullscreen = value;
                        OnFullscreenChanged(EventArgs.Empty);
                    });

                    if (InvokeRequired)
                        BeginInvoke(action);
                    else
                        action();
                }
            }
        }

        protected GameWindowBase(bool fullscreen = false, bool disableCloseButton = true)
        {
            SetStyle(ControlStyles.UserPaint, true);
            UpdateStyles();

            this.fullscreen = fullscreen;
            this.disableCloseButton = disableCloseButton;
            if (fullscreen)
            {
                oldBorderStyle = FormBorderStyle.Sizable;
                oldWindowState = FormWindowState.Normal;

                FormBorderStyle = FormBorderStyle.None;
                WindowState = FormWindowState.Maximized;
            }
        }

        protected virtual void OnFullscreenChanged(EventArgs e)
        {
            if (fullscreen)
            {
                oldBorderStyle = FormBorderStyle;
                oldWindowState = WindowState;

                FormBorderStyle = FormBorderStyle.None;
                WindowState = FormWindowState.Maximized;
            }
            else
            {
                FormBorderStyle = oldBorderStyle;
                WindowState = oldWindowState;
            }

            if (FullscreenChanged != null)
                FullscreenChanged(this, e);
        }

        protected override void OnPaint(PaintEventArgs e)
        { }

        protected override void OnPaintBackground(PaintEventArgs e)
        { }
    }
}
