﻿using System;
using GameUtils.Graphics.Data;
using GameUtils.Math;

namespace GameUtils.Graphics
{
    public class LinearGradientBrush : Brush<LinearGradientBrushData>
    {
        GradientStop[] gradientStops;
        Vector2 startPoint;
        Vector2 endPoint;
        bool stopsChanged;
        bool startChanged;
        bool endChanged;
        readonly object locker;

        /// <summary>
        /// A list of gradient stops defining the colors with their positions in the gradient.
        /// </summary>
        public GradientStop[] GradientStops
        {
            get { return gradientStops; }
            set
            {
                if (value == null)
                    throw new ArgumentNullException("value");
                if (value.Length < 2)
                    throw new ArgumentException("Es müssen mindestens zwei Verlaufsfarben angegeben sein.", "value");

                lock (locker)
                {
                    gradientStops = value;
                    stopsChanged = true;
                }
            }
        }

        /// <summary>
        /// The starting point of the gradient.
        /// </summary>
        public Vector2 StartPoint
        {
            get { return startPoint; }
            set
            {
                lock (locker)
                {
                    startPoint = value;
                    startChanged = true;
                }
            }
        }

        /// <summary>
        /// The ending point of the gradient.
        /// </summary>
        public Vector2 EndPoint
        {
            get { return endPoint; }
            set
            {
                lock (locker)
                {
                    endPoint = value;
                    endChanged = true;
                }
            }
        }

        /// <summary>
        /// Cretes a new linear gradient brush.
        /// </summary>
        public LinearGradientBrush(GradientStop[] stops, Vector2 startPoint, Vector2 endPoint)
        {
            if (stops == null)
                throw new ArgumentNullException("stops");
            if (stops.Length < 2)
                throw new ArgumentException("Es müssen mindestens zwei Verlaufsfarben angegeben sein.", "stops");

            gradientStops = stops;
            this.startPoint = startPoint;
            this.endPoint = endPoint;
            locker = new object();
        }

        protected override void ApplyChanges()
        {
            lock (locker)
            {
                if (stopsChanged)
                {
                    Data.GradientStops = gradientStops;
                    stopsChanged = false;
                }
                if (startChanged)
                {
                    Data.StartPoint = startPoint;
                    startChanged = false;
                }
                if (endChanged)
                {
                    Data.EndPoint = endPoint;
                    endChanged = false;
                }
            }

            base.ApplyChanges();
        }

        protected override LinearGradientBrushData CreateData(Factory factory)
        {
            lock (locker)
            {
                stopsChanged = false;
                startChanged = false;
                endChanged = false;
                return factory.CreateLinearGradientBrush(gradientStops, startPoint, endPoint);
            }
        }
    }
}
